﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Windows.Forms;
using Microsoft.Xna.Framework;

namespace HumanMotionSolver
{
    public partial class FrmMain : Form
    {
        // иерархия сочленений выполнена в виде линейного массива с индексами родителей
        // если индекс родителя -1, то это корневой элемент, он нулевой в масиве
        Joint[] _hierarchy;

        // кадры анимации с расчитанными положениями сочленений
        AnimationFrame[] _frames;

        // разделители значений в документе BVH
        readonly char[] _splitchars = new[] { ' ', '\t' };

        // корневой элемент дерева скелета в TreeView
        readonly TreeNode _modelNode = new TreeNode("Модель");

        // угол поворота камеры по горизонтали
        float _angle = MathHelper.PiOver4;
        // угол поворота камеры по вертикали
        float _vangle = MathHelper.PiOver4;
        // текущий кадр анимации
        int _frameNum;

        // удаленность камеры от модели
        float _scale = 1f;

        // флаги нажатия кнопок мыши
        bool _leftMouseDown;
        bool _rightMouseDown;

        // положение мыши с прошлого перемещения
        System.Drawing.Point _point = System.Drawing.Point.Empty;

        // конструктор формы
        public FrmMain()
        {
            InitializeComponent();
            treeView1.Nodes.Add(_modelNode);
        }


        // функция открытия файла
        private void MenuOpenClick(object sender, EventArgs e)
        {
            _frameNum = 0;
            OpenFileDialog ofd = new OpenFileDialog {Filter = "Файл BVH|*.bvh"};
            string[] fileData;

            // очистка дерева модели
            _modelNode.Nodes.Clear();

            // диалог открытия
            if (ofd.ShowDialog(this) == DialogResult.OK)
            {
                // чтение всех строк фала
                fileData = File.ReadAllLines(ofd.FileName);

                if (fileData.Length > 0)
                {
                    // чтение иерархии скелета
                    ReadHierarchy(ref fileData);
                    // чтение анимации
                    ReadMotion(ref fileData);
                    // раскрытие дерева скелета в TreeView
                    _modelNode.ExpandAll();
                }
                else
                {
                    MessageBox.Show(this, "Файл не содержит данных!", "Ошибка чтения файла!",
                        MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }

        }

        // функция чтения иерархии сочленений
        private void ReadHierarchy(ref string[] fileData)
        {
            // список сочленений
            List<Joint> joints = new List<Joint>();

            // инициализация переменных, необходимых для чтения иерархии сочленений скелета
            int parent = -1;
            Joint joint = null;
            string line;
            string name = "";
            TreeNode node = _modelNode;

            for (int i = 0; i < fileData.Length; i++)
            {
                // поиск строки с ключевым словом HIERARCHY
                if (fileData[i].Trim().ToUpper().StartsWith("HIERARCHY", StringComparison.InvariantCultureIgnoreCase))
                {
                    int current = i;

                    // чтение иерархии
                    while (current < fileData.Length - 1)
                    {
                        current++;

                        // текущая строка
                        line = fileData[current].Trim();

                        // пропускаем пустые строки, если таковые есть
                        if (line == "")
                            continue;

                        // если строка начинается с "{"
                        if (line.StartsWith("{", StringComparison.InvariantCultureIgnoreCase))
                        {
                            //создаем новое сочленение и делаем его текущим
                            joint = new Joint {Name = name, Parent = parent};
                            joints.Add(joint);
                            parent = joints.Count - 1;

                            // создаем узел дерева
                            TreeNode parentNode = node;
                            node = new TreeNode(name);
                            parentNode.Nodes.Add(node);
                        }
                        // если строка начинается с "}"
                        else if (line.StartsWith("}", StringComparison.InvariantCultureIgnoreCase))
                        {
                            // возвращение на предидущий 
                            parent = joints[parent].Parent;
                            node = node.Parent;

                            // если родитель -1 то останавливаем чтение иерархии
                            if (parent == -1)
                                break;
                        }
                        // если строка начинается с "OFFSET"
                        else if (line.StartsWith("OFFSET", StringComparison.InvariantCultureIgnoreCase))
                        {
                            string[] vals = line.Split(_splitchars, StringSplitOptions.RemoveEmptyEntries);

                            // чтение смещения сочленения
                            if (joint != null)
                            {
                                joint.Offset.X = Val(vals[1]);
                                joint.Offset.Y = Val(vals[2]);
                                joint.Offset.Z = Val(vals[3]);

                                // установка имени сочленения в дереве с учетом позиции
                                node.Text = joint.ToString();
                            }
                        }
                        // если строка начинается с "CHANNELS"
                        else if (line.StartsWith("CHANNELS", StringComparison.InvariantCultureIgnoreCase))
                        {
                            string[] vals = line.Split(_splitchars, StringSplitOptions.RemoveEmptyEntries);
                            // чтение количества каналов
                            if (joint != null)
                            {
                                joint.CountChennels = int.Parse(vals[1]);
                                // чтение значений каналов в массив сочленения
                                joint.Chennels = new string[joint.CountChennels];
                                Array.Copy(vals, 2, joint.Chennels, 0, joint.CountChennels);
                            }
                        }
                        // если строка начинается с "JOINT" или "ROOT"
                        else if (line.StartsWith("JOINT", StringComparison.InvariantCultureIgnoreCase) ||
                            line.StartsWith("ROOT", StringComparison.InvariantCultureIgnoreCase))
                        {
                            // читаем имя сочленения
                            name = line.Split(_splitchars, StringSplitOptions.RemoveEmptyEntries)[1];
                        }
                        // если строка начинается с "End Site"
                        else if (line.StartsWith("End Site", StringComparison.InvariantCultureIgnoreCase))
                        {
                            name = "End";
                        }
                        // иначе заканчиваем чтение, если попадается что-то другое
                        else
                        {
                            break;
                        }
                    }
                    break;
                }
            }
            // записываем сочленение в массив _hierarchy
            _hierarchy = joints.ToArray();
        }


        // функция чтения кадров анимации
        private void ReadMotion(ref string[] fileData)
        {
            for (int i = 0; i < fileData.Length; i++)
            {
                // ищем строку, содержащую ключевое слово MOTION
                if (fileData[i].Trim().ToUpper().StartsWith("MOTION", StringComparison.InvariantCultureIgnoreCase))
                {
                    // вспомогательные переменные

                    // чтение числа кадров
                    string line = fileData[i + 1].Trim();
                    string[] vals = line.Split(_splitchars, StringSplitOptions.RemoveEmptyEntries);
                    int countFrames = int.Parse(vals[vals.Length - 1]);

                    // чтение скорости анимации
                    line = fileData[i + 2].Trim();
                    vals = line.Split(_splitchars, StringSplitOptions.RemoveEmptyEntries);
                    float speed = Val(vals[vals.Length - 1]);

                    // установка интервала таймера
                    FrameTimer.Interval = (int)(speed * 1000);

                    // списки для временного хранения данных
                    List<float> data = new List<float>();
                    List<Matrix> transforms = new List<Matrix>();
                    List<Vector3> positions = new List<Vector3>();

                    // создаем новый массив кадров
                    _frames = new AnimationFrame[countFrames];


                    // чтение кадров анимации
                    for (int j = 0; j < countFrames; j++)
                    {
                        // индекс текущей строки для чтения
                        int index = j + i + 3;

                        data.Clear();
                        positions.Clear();
                        transforms.Clear();

                        _frames[j] = new AnimationFrame();

                        // чтение строки с данными
                        line = fileData[index].Trim();
                        // разделение строки на компоненты
                        vals = line.Split(_splitchars, StringSplitOptions.RemoveEmptyEntries);

                        // чтение значений компонент
                        for (int k = 0; k < vals.Length; k++)
                        {
                            data.Add(Val(vals[k]));
                        }

                        // смещение, относительно начала строки данных
                        int dataIndex = 0;

                        // чтение каналов трансформаций
                        for (int k = 0; k < _hierarchy.Length; k++)
                        {
                            Matrix transform = Matrix.Identity;

                            // применяем трансформации в порядке, указанном в BVH afqkt
                            for (int l = 0; l < _hierarchy[k].CountChennels; l++)
                            {
                                switch (_hierarchy[k].Chennels[l].ToLower().Trim())
                                {
                                    case "xrotation":
                                        transform *= Matrix.CreateRotationX(MathHelper.ToRadians(data[dataIndex + l]));
                                        break;
                                    case "yrotation":
                                        transform *= Matrix.CreateRotationY(MathHelper.ToRadians(data[dataIndex + l]));
                                        break;
                                    case "zrotation":
                                        transform *= Matrix.CreateRotationZ(MathHelper.ToRadians(data[dataIndex + l]));
                                        break;
                                    // позицию игнорируем в текущей задаче
                                    //case "xposition":
                                    //    transform *= Matrix.CreateTranslation(new Vector3(data[dataIndex + l], 0, 0));
                                    //    break;
                                    //case "yposition":
                                    //    transform *= Matrix.CreateTranslation(new Vector3(0, data[dataIndex + l], 0));
                                    //    break;
                                    //case "zposition":
                                    //    transform *= Matrix.CreateTranslation(new Vector3(0, 0, data[dataIndex + l]));
                                    //    break;
                                }
                            }

                            // увеличиваем смещение на количество каналов анимации в данном сочленении
                            dataIndex += _hierarchy[k].CountChennels;

                            // сохраняем полученную трансформацию
                            transforms.Add(transform);
                        }

                        // расчет результирующих трансформаций

                        // первая трансформация не имеет ролителя
                        transforms[0] = transforms[0] * Matrix.CreateTranslation(_hierarchy[0].Offset);

                        // остальные трансформации
                        for (int k = 1; k < transforms.Count; k++)
                        {
                            transforms[k] = transforms[k] * Matrix.CreateTranslation(_hierarchy[k].Offset) *
                                transforms[_hierarchy[k].Parent];
                        }

                        // получение координат сочленений
                        for (int k = 0; k < transforms.Count; k++)
                        {
                            positions.Add(transforms[k].Translation);
                        }

                        // сохранение результатов кадра
                        _frames[j].Positions = positions.ToArray();

                    }
                }
            }
        }

        // функция отрисовки скелета
        private void CanvasPaint(object sender, PaintEventArgs e)
        {
            Graphics g = e.Graphics;

            // очистка области рисования
            g.Clear(Color.White);

            // если загружена иерархия то выполняем отрисовку
            if (_hierarchy != null)
            {
                // временный массив сочленений
                Joint[] points = new Joint[_hierarchy.Length];

                // переключение состояний
                if (RadioStartPos.Checked) // начальная позиция
                {
                    // копируем дерево сочленений
                    for (int i = 0; i < points.Length; i++)
                    {
                        points[i] = new Joint
                                        {
                                            Offset = _hierarchy[i].Offset, 
                                            Parent = _hierarchy[i].Parent
                                        };
                    }
                    // рассчитываем абсолютные положения
                    for (int i = 1; i < points.Length; i++)
                    {
                        points[i].Offset += points[points[i].Parent].Offset;
                    }
                }
                else // анимация
                {
                    // если кадр анимации превышает длину анимации то проигрываем анимацию сначала
                    if (_frameNum >= _frames.Length)
                    {
                        _frameNum = 0;
                    }

                    // копирование рассчитанных позиций из массива кадров
                    for (int i = 0; i < points.Length; i++)
                    {
                        points[i] = new Joint
                                        {
                                            Offset = _frames[_frameNum].Positions[i],
                                            Parent = _hierarchy[i].Parent
                                        };
                    }

                    // увеличение кадра на 1
                    _frameNum++;
                }

                // видовая матрица
                Matrix view = Matrix.CreatePerspectiveFieldOfView((float)Math.PI / 4f,
                    Canvas.Width / (float)Canvas.Height, 1f, 1000);

                // расчет матрицы трансформации точки взгляда камеры
                Matrix rot = Matrix.CreateRotationZ(_vangle);
                rot *= Matrix.CreateRotationY(_angle);


                // матрица проекций 
                Matrix projection = Matrix.CreateLookAt(
                    Vector3.Transform(new Vector3(250, 0, 0), rot) * _scale, 
                    _hierarchy[0].Offset, Vector3.Up);

                // результирующая матрица перобразования
                Matrix vp = projection * view;


                // отрисовка линий сетки
                for (int i = 0; i < 11; i++)
                {
                    // расчет координат в экранном пространстве
                    Vector4 pos1 = Vector4.Transform(new Vector3(-50, 0, -50 + i * 10), vp);
                    Vector4 pos2 = Vector4.Transform(new Vector3(50, 0, -50 + i * 10), vp);

                    // четвертая компонента вектора используется для нормализации координат
                    pos1 /= pos1.W;
                    pos2 /= pos2.W;

                    // пересчет позиции из экранного пространства в окно проекции
                    pos1.X = ((1f - pos1.X) / 2f) * Canvas.Width;
                    pos1.Y = ((1f - pos1.Y) / 2f) * Canvas.Height;
                    pos2.X = ((1f - pos2.X) / 2f) * Canvas.Width;
                    pos2.Y = ((1f - pos2.Y) / 2f) * Canvas.Height;

                    // отрисовка линии
                    g.DrawLine(Pens.LightGray, new PointF(pos1.X, pos1.Y), new PointF(pos2.X, pos2.Y));

                    // поаторяем для перпендикулярной линии
                    pos1 = Vector4.Transform(new Vector3(-50 + i * 10, 0, -50), vp);
                    pos2 = Vector4.Transform(new Vector3(-50 + i * 10, 0, 50), vp);

                    pos1 /= pos1.W;
                    pos2 /= pos2.W;

                    pos1.X = ((1f - pos1.X) / 2f) * Canvas.Width;
                    pos1.Y = ((1f - pos1.Y) / 2f) * Canvas.Height;
                    pos2.X = ((1f - pos2.X) / 2f) * Canvas.Width;
                    pos2.Y = ((1f - pos2.Y) / 2f) * Canvas.Height;

                    g.DrawLine(Pens.LightGray, new PointF(pos1.X, pos1.Y), new PointF(pos2.X, pos2.Y));
                }


                // отрисовка скелета
                for (int i = 0; i < points.Length; i++)
                {
                    // расчет координаты в экранном пространстве
                    Vector4 pos = Vector4.Transform(points[i].Offset, vp);

                    // четвертая компонента вектора используется для нормализации координат
                    pos /= pos.W;

                    // пересчет позиции из экранного пространства в окно проекции
                    // (позицию сохраняем в массиве)
                    points[i].Offset = new Vector3(pos.X, pos.Y, pos.Z);
                    points[i].Offset.X = ((1f - points[i].Offset.X) / 2f) * Canvas.Width;
                    points[i].Offset.Y = ((1f - points[i].Offset.Y) / 2f) * Canvas.Height;

                    // если сочленение имеет родителя то рисуем линию из родительского сочленения в данное
                    if (points[i].Parent != -1)
                    {
                        g.DrawLine(Pens.Gray, new PointF(points[i].Offset.X, points[i].Offset.Y),
                            new PointF(points[points[i].Parent].Offset.X, points[points[i].Parent].Offset.Y));
                    }

                    // вывод точки сочленения
                    g.FillEllipse(Brushes.Red, new RectangleF(points[i].Offset.X - 2f, 
                        points[i].Offset.Y - 2f, 4f, 4f));
                }
            }
        }

        // при срабатывании таймера вызывается перерисовка
        private void FrameTimerTick(object sender, EventArgs e)
        {
            Canvas.Invalidate();
        }

        // переключение состояний
        private void RadioCheckedChanged(object sender, EventArgs e)
        {
            _frameNum = 0;
        }

        // обработка нажатия мыши
        private void CanvasMouseDown(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                _leftMouseDown = true;

            }
            else if (e.Button == MouseButtons.Right)
            {
                _rightMouseDown = true;

            }
        }

        // обработка перемещения мыши
        private void CanvasMouseMove(object sender, MouseEventArgs e)
        {
            // новая позиция мыши
            System.Drawing.Point newPoint = Canvas.PointToClient(Cursor.Position);

            // углы поворота
            if (_leftMouseDown)
            {
                _angle += 0.05f * (newPoint.X - _point.X);
                _vangle += 0.05f * (newPoint.Y - _point.Y);
                if (_vangle <= -MathHelper.Pi / 2.1f)
                {
                    _vangle = -MathHelper.Pi / 2.1f;
                }

                if (_vangle >= MathHelper.Pi / 2.1f)
                {
                    _vangle = MathHelper.Pi / 2.1f;
                }
            }

            // масштаб
            if (_rightMouseDown)
            {
                _scale += 0.05f * System.Drawing.Point.Subtract(newPoint, new Size(_point)).Y;
                if (_scale <= 0.5f)
                {
                    _scale = 0.5f;
                }

                if (_scale >= 4f)
                {
                    _scale = 4f;
                }
            }

            _point = Canvas.PointToClient(Cursor.Position);
        }

        // обработка отжатия мыши
        private void CanvasMouseUp(object sender, MouseEventArgs e)
        {
            switch (e.Button)
            {
                case MouseButtons.Left:
                    _leftMouseDown = false;
                    break;
                case MouseButtons.Right:
                    _rightMouseDown = false;
                    break;
            }
        }

        // вспомогательная функция преобразования строки в вещественное число с учетом текущей культуры
        private static float Val(string value)
        {
            string sep = System.Globalization.CultureInfo.CurrentCulture.NumberFormat.NumberDecimalSeparator;
            return float.Parse(value.Replace(".", sep));
        }

    }
}
